package de.fraunhofer.iem.dataprovider.taskManager.tasks

import de.fraunhofer.iem.dataprovider.sarif.Sarif
import de.fraunhofer.iem.dataprovider.sarif.getSarifFromFilePath
import de.fraunhofer.iem.dataprovider.taskManager.SarifProcessDone
import de.fraunhofer.iem.dataprovider.taskManager.SarifProcessGroupDone
import de.fraunhofer.iem.dataprovider.toolResult.ToolRunService
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import java.io.IOException
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.util.*

sealed class ProcessTask : Task() {
    protected abstract val flags: Array<String>
    protected abstract val execPath: String


    private val mainScope = CoroutineScope(Dispatchers.Default)
    override suspend fun execute() {
        if (isExecutable(execPath)) {
            val process = ProcessBuilder(execPath, *flags).start()

            process.onExit().thenApply { p1 ->
                mainScope.launch {
                    handleProcessReturn(p1)
                }
            }
        } else {
            logger.warn("Given execPath is not an executable $execPath.")
            // TODO: we should probably throw an exception here.
        }
    }

    private fun isExecutable(filePath: String): Boolean {
        val path = Paths.get(filePath)
        return Files.isExecutable(path)
    }

    abstract suspend fun handleProcessReturn(p: Process)
    abstract suspend fun sendResult(sarif: Sarif)

}

sealed class SarifTask : ProcessTask() {
    protected abstract val repoId: UUID
    protected abstract val toolRunService: ToolRunService
    protected abstract val resultPath: Path


    override suspend fun handleProcessReturn(p: Process) {

        // TODO: check process exit codes. Current problem, if an analysis
        //  has findings it might return an exit code != 0 even tho the process finished correctly

        logger.info("Handle Process return in $javaClass")
        try {
            val sarifResult = getSarifFromFilePath(resultPath)
            toolRunService.saveSarif(sarifResult, repoId)
            sendResult(sarifResult)
        } catch (e: IOException) {
            // TODO: send meaningful event and react on error
        }
    }

    override suspend fun sendResult(sarif: Sarif) {

        val event = if (groupID != null) {
            SarifProcessGroupDone(this.taskID, repoId, groupID!!, sarif)
        } else {
            SarifProcessDone(this.taskID, repoId, sarif)
        }

        responseChannel(event)
    }
}
