package de.fraunhofer.iem.dataprovider.toolResult

import de.fraunhofer.iem.dataprovider.gitlab.Repository
import jakarta.persistence.*
import org.hibernate.Hibernate
import org.hibernate.annotations.JdbcTypeCode
import org.hibernate.type.SqlTypes
import java.sql.Timestamp
import java.util.*

enum class Level {
    Warning, Error, Note, None
}

enum class Kind {
    Pass, Open, Informational, NotApplicable
}

@Entity
@Table(name = "tool_run")
class ToolRun {
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "id", nullable = false)
    var id: UUID? = null

    @OneToOne(orphanRemoval = true)
    @JoinColumn(name = "tool_id")
    var tool: Tool? = null

    @JdbcTypeCode(SqlTypes.TIMESTAMP)
    @Column(name = "time_stamp")
    var timeStamp: Timestamp? = null

    @ManyToOne(fetch = FetchType.EAGER)
    var repository: Repository? = null

    @OneToMany(
        mappedBy = "toolRun",
        orphanRemoval = true
    )
    var rules: MutableSet<Rule> = mutableSetOf()

    @OneToMany(mappedBy = "toolRun", cascade = [CascadeType.ALL], orphanRemoval = true)
    var findings: MutableSet<Finding> = mutableSetOf()

    fun addFindings(findings: Collection<Finding>) {
        findings.forEach {
            this.addFinding(it)
        }
    }

    fun addFinding(finding: Finding) {
        findings.add(finding)
        finding.toolRun = this
    }

    fun addRules(rules: Collection<Rule>) {
        rules.forEach {
            this.addRule(it)
        }
    }

    fun addRule(rule: Rule) {
        rules.add(rule)
        rule.toolRun = this
    }
}

@Entity
@Table(name = "tool")
class Tool {
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "id", nullable = false)
    var id: UUID? = null

    @Column(name = "full_name", length = 500)
    var fullName: String? = null

    @Column(name = "name", length = 500)
    var name: String? = null

    @Column(name = "version")
    var version: String? = null

    @OneToMany(
        mappedBy = "tool",
        cascade = [CascadeType.MERGE, CascadeType.REFRESH, CascadeType.REMOVE, CascadeType.DETACH],
        orphanRemoval = true
    )
    var rules: MutableSet<Rule> = mutableSetOf()

    fun addRules(rules: Collection<Rule>) {
        rules.forEach {
            this.addRule(it)
        }
    }

    fun addRule(rule: Rule) {
        rules.add(rule)
        rule.tool = this
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other == null || Hibernate.getClass(this) != Hibernate.getClass(other)) return false
        other as Tool

        return id != null && id == other.id
    }

    override fun hashCode(): Int = javaClass.hashCode()
}


@Entity
@Table(name = "finding")
class Finding {
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "id", nullable = false)
    var id: UUID? = null

    @ManyToOne
    var rule: Rule? = null

    @OneToMany(mappedBy = "finding", cascade = [CascadeType.ALL], orphanRemoval = true)
    var locations: MutableSet<Location> = mutableSetOf()

    fun addLocations(locations: Collection<Location>) {
        locations.forEach {
            this.addLocation(it)
        }
    }

    fun addLocation(location: Location) {
        locations.add(location)
        location.finding = this
    }

    @Enumerated
    @Column(name = "level")
    var level: Level? = null

    @Enumerated
    @Column(name = "kind")
    var kind: Kind? = null

    @Column(name = "message", columnDefinition = "TEXT")
    var message: String? = null

    @ManyToOne
    var toolRun: ToolRun? = null

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other == null || Hibernate.getClass(this) != Hibernate.getClass(other)) return false
        other as Finding

        return id != null && id == other.id
    }

    override fun hashCode(): Int = javaClass.hashCode()
}

@Entity
@Table(name = "location")
class Location {
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "id", nullable = false)
    var id: UUID? = null

    @ManyToOne
    var finding: Finding? = null

    @Column(name = "artifact_location", columnDefinition = "TEXT")
    var artifactLocation: String? = null

    @Column(name = "end_column")
    var endColumn: Int? = null

    @Column(name = "end_line")
    var endLine: Int? = null

    @Column(name = "start_column")
    var startColumn: Int? = null

    @Column(name = "start_line")
    var startLine: Int? = null

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other == null || Hibernate.getClass(this) != Hibernate.getClass(other)) return false
        other as Location

        return id != null && id == other.id
    }

    override fun hashCode(): Int = javaClass.hashCode()
}

@Entity
@Table(name = "rule")
class Rule {
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "id", nullable = false)
    var id: UUID? = null


    @Column(name = "sarif_rule_id")
    var sarifRuleId: String? = null

    @Column(name = "short_description", columnDefinition = "TEXT")
    var shortDescription: String? = null

    @OneToMany(
        mappedBy = "rule",
        cascade = [CascadeType.MERGE, CascadeType.REFRESH, CascadeType.REFRESH, CascadeType.DETACH],
        orphanRemoval = true
    )
    var findings: MutableSet<Finding> = mutableSetOf()

    @ManyToOne
    var toolRun: ToolRun? = null

    fun addResultLocations(findings: Collection<Finding>) {
        findings.forEach {
            this.addResultLocation(it)
        }
    }

    fun addResultLocation(finding: Finding) {
        findings.add(finding)
        finding.rule = this
    }

    @ElementCollection
    @CollectionTable(name = "rule_messages", joinColumns = [JoinColumn(name = "owner_id")])
    @Column(name = "message", length = 1000)
    var messages: MutableList<String> = mutableListOf()

    @ManyToOne
    var tool: Tool? = null

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other == null || Hibernate.getClass(this) != Hibernate.getClass(other)) return false
        other as Rule

        return id != null && id == other.id
    }

    override fun hashCode(): Int = javaClass.hashCode()
}